#!/usr/bin/env python3
#
#  This file is part of ProfileView.
#
#  ProfileView is free software: you can redistribute it and/or modify
#  it under the terms of the CeCILL 2.1 Licence
#
#  ProfileView is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  You should have received a copy of the Licence CeCILL 2.1 along
#  with ProfileView. If not, see <https://cecill.info/>.
#

import sys, os, argparse
import time
import re
import math
import pickle, gzip

from pv_utils import *
from hh_utils import hhm, parse_hhm

script_name = re.sub( r'\.py$', '', os.path.basename(__file__) )

def main(argv=None):
    start = time.perf_counter()

    parser = argparse.ArgumentParser()
    parser.add_argument('--hhm-file', dest='hhmFile', type=str, help='hhm file path')
    parser.add_argument('--hhm-dir', dest='hhmDir', type=str, help='hhm directory path')
    parser.add_argument('--prefix', dest='outPrefix', default="hhdict", type=str, help='output prefix of generated dictionary')
    args = parser.parse_args()

    # parameters check
    validParameters = True;
    if args.hhmFile is None and args.hhmDir is None:
        validParameters = False
        print_error('either one of --hhm-file or --hhm-dir arguments must be set')
    if args.hhmFile is not None and args.hhmDir is not None:
        validParameters = False
        print_error('it is not possible to use both --hhm-file and --hhm-dir arguments')
    elif args.hhmFile is not None and not os.path.isfile(args.hhmFile):
        validParameters = False
        print_error('--hhm-file \"{}\" is not a file'.format(args.hhmFile))
    elif args.hhmDir is not None and not os.path.isdir(args.hhmDir):
        validParameters = False
        print_error('--hhm-dir \"{}\" is not a directory'.format(args.hhmDir))
    if not validParameters:
        return 1

    fileList = []
    if args.hhmFile is not None:
        fileList = [ args.hhmFile ]
    if args.hhmDir is not None:
        hhms_dir = os.path.abspath(args.hhmDir)
        fileList = [ '{}/{}'.format(hhms_dir,f) for f in os.listdir(hhms_dir) if f.endswith('.hhm') and os.path.isfile('{}/{}'.format(hhms_dir,f)) ]

    if len(fileList) == 0:
        print_warning('no hhm file found in directory "{}"'.format(args.hhmDir))

    print_status("parsing HHM files")
    hhmDict = {}
    for fname in fileList:
        with open(fname,'r') as fh:
            hhm = parse_hhm(fh)
            hhmDict[ hhm.name ] = hhm

    hhmDictFileName = "{}.pgz".format(args.outPrefix)
    print_status("writing hhm dictionary to \"{}\"".format(script_name,hhmDictFileName))
    with gzip.GzipFile(hhmDictFileName, 'w') as hhmDictFile:
        pickle.dump( hhmDict, hhmDictFile )

    exec_time = time.perf_counter() - start
    print_status('execution time: {:.2f}s'.format(exec_time));

    return 0

# Check if the program is not being imported
if __name__ == "__main__":
    sys.exit(main())
