#!/ usr/bin/env python
#
#  This file is part of ProfileView.
#
#  ProfileView is free software: you can redistribute it and/or modify
#  it under the terms of the CeCILL 2.1 Licence
#
#  ProfileView is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  You should have received a copy of the Licence CeCILL 2.1 along
#  with ProfileView. If not, see <https://cecill.info/>.
#

import sys, errno, os
import argparse
import gzip
import re
import pickle

from hmm_utils import *
from pv_utils import *


def parseHMMER(fh,hmmDict):
    hmm_name = ""
    seq_name = ""
    line = fh.readline()
    while line:
        tokens = line.split()
        if len(tokens) == 0 or line.startswith("#"):
            pass
        elif line.startswith("Query: "):
            hmm_name = tokens[1]
        elif line.startswith(">> "): # target sequence section
            seq_name = tokens[1]
            # skip seqName domain table header
            fh.readline()
            fh.readline()
            # process domain table
            line = fh.readline()
            hitTable = []
            while line and not line.lstrip().startswith("== domain ") and not line.startswith(">> ") and not line.startswith("Query: "):
                tokens = line.split()
                ## domain table row:
                ## domNum  [!?]  score  bias  c-Evalue  i-Evalue  hmmfrom  hmm-to  .. alifrom  ali-to  ..  envfrom  env-to  ..  acc
                if len(tokens) == 16:
                    #eprint('[debug] domain-table: {}'.format(line[:-1]))
                    hit = hmmer_hit()
                    hit.hmm_name = hmm_name
                    hit.seq_name = seq_name
                    hit.typ = tokens[1]
                    hit.score = float(tokens[2])
                    hit.hmm_from = int(tokens[6])
                    hit.hmm_to   = int(tokens[7])
                    hit.ali_from = int(tokens[9])
                    hit.ali_to   = int(tokens[10])
                    hit.env_from = int(tokens[12])
                    hit.env_to   = int(tokens[13])
                    hitTable.append(hit)
                line = fh.readline()
            # process each domain hit (i.e., hmm-seq alignment)
            while line and line.lstrip().startswith("== domain "):
                tokens = line.split()
                hit_index = int(tokens[2])-1
                hit = hitTable[hit_index]
                hmm_cons = ""
                seq_cons = ""
                pp_cons  = ""
                aln_to = 0
                while True:
                    line = fh.readline()
                    tokens = line.split()
                    if len(tokens) == 4 and tokens[0] == hit.hmm_name:
                        hmm_cons += tokens[2]
                        #fh.readline() # skip hmm-seq match string
                        match_cons_i = re.search(r"\s*\S+\s+[0-9]+",line).end()+1
                        match_line   = fh.readline()
                        hit.match_str += match_line[ match_cons_i : match_cons_i+len(tokens[2]) ]
                        tokens = fh.readline().split()
                        seq_cons += tokens[2]
                        if tokens[3].isdigit():
                            aln_to = int(tokens[3])
                        tokens = fh.readline().split()
                        pp_cons += tokens[0]
                        if aln_to == hit.ali_to:
                            break
                hit.hmm_cons = hmm_cons
                hit.seq_cons = seq_cons
                hit.pp_cons = pp_cons
                if hit.typ == '!': # significant hit
                    hmm = hmmDict[hit.hmm_name]
                    # compute scores
                    hit.aligned_cols = 0
                    hit.well_aligned_cols = 0
                    hit.matches = 0
                    hit.wcscore = 0.0
                    hit.mcscore = 0.0
                    hmm_i = hit.hmm_from
                    for i,hmm_c in enumerate(hmm_cons):
                        hit.ali_len += 1
                        if hmm_c == '.':  # insertion in sequence w.r.t. the model
                            continue
                        seq_c = seq_cons[i]
                        if seq_c == '-':  # deletion in sequence w.r.t. the model
                            hmm_i += 1
                            continue
                        hit.aligned_cols += 1
                        hit.matches += 1 if hmm_c.upper() == seq_c.upper() else 0
                        if seq_c in 'ACDEFGHIKLMNPQRSTVWY' and pp_cons[i] in list('89*'):
                            score_i = hmm.score(hmm_i,seq_c)
                            hit.well_aligned_cols += 1
                            if hmm_c.upper() == seq_c.upper():
                                hit.cm_str += hit.match_str[i]
                                hit.mcscore += score_i
                            if score_i >= 3.0: # high-score position
                                hit.wc_str += hit.match_str[i]
                                hit.wcscore += score_i
                        hmm_i += 1
                    yield hit
                # skip lines until the next domain/sequence or the end of file
                line = fh.readline()
                while line and not line.lstrip().startswith("== domain ") and not line.startswith(">> ") and not line.startswith("Query: "):
                    line = fh.readline()
            # here either EOF is reached or there are other models/sequences to process
            continue
        line = fh.readline()


def main( argv = None ):

    # parameter definition
    parser = argparse.ArgumentParser()
    parser.add_argument('--hmmer-dict', dest='hmmerDictFile', type=str, required=True, help='HMMER dictionary .pgz file')
    parser.add_argument('--seq-db', dest='seqFile', type=str, required=True, help='Sequence database in csv format with [name,function,family,length] fields')
    args = parser.parse_args()

    # parameter validation
    validParameters = True;
    if not os.path.isfile(args.hmmerDictFile):
        validParameters = False
        print_error('HMMER dictionary file "{}" does not exist'.format(args.hmmerDictFile))
    if not os.path.isfile(args.seqFile):
        validParameters = False
        print_error('Sequence database file "{}" does not exist'.format(args.seqFile))
    if not validParameters:
        return 1

    print_status("loading HMMER dictionary...")
    hmmerDict = {}
    with gzip.open(args.hmmerDictFile,'rb') as f:
        hmmerDict = pickle.load(f)

    print_status("loading sequence database...")
    seqDict = {}
    with open( args.seqFile, mode='r', newline=None ) as seqFile:
        for line in seqFile:
            line = line.strip()
            if line:
                seq_name, seq_fun, seq_fam, seq_len = [ x.strip() for x in line.split(',') ]
                seqDict[ seq_name ] = { 'Function':seq_fun, 'Family':seq_fam, 'Length':int(seq_len) }

    print_status("processing hmmsearch output files...")

    header = ['#hmm_name','hmm_len','seq_name','seq_len','seq_family','seq_func',
            'bitscore', 'mean_score', 'mcscore','mean_mcs', 'wcscore','mean_wcs',
            'ident','hmm_cov','hit_type']
    print('\t'.join(header))
    record = []
    prev_hit = hmmer_hit()
    for hit in parseHMMER(sys.stdin,hmmerDict):
        if hit.seq_name not in seqDict:
            continue
        hmm = hmmerDict[hit.hmm_name]
        if prev_hit.canExtendWith(hit):
            hit = prev_hit.extendWith(hit)
        elif len(record) > 0:
            print('\t'.join(record))
            record = []

        hmm_len = hmm.length
        hmm_reg = hit.hmm_to-hit.hmm_from+1
        hmm_loh = hit.hmm_from-1 # left overhang of the hmm hit
        hmm_roh = hmm_len-hit.hmm_to # right overhang of the hmm hit
        hmm_cov = 100.0 * hmm_reg / hmm_len

        seq_name = hit.seq_name
        seq_len  = seqDict[seq_name]['Length']
        seq_loh  = hit.env_from-1 # left overhang of the seq hit (envelope considered)
        seq_roh  = seq_len-hit.env_to # right overhang of the seq hit (envelope considered)

        max_oh   = max( 0.1 * hmm_len, 10 )
        #left_oh  = min(hmm_loh,seq_loh)
        #right_oh = min(hmm_roh,seq_roh)
        hit_type = "FULL" if (hmm_loh <= max_oh and hmm_roh <= max_oh) else ("OVER" if (hmm_loh-seq_loh > max_oh or hmm_roh-seq_roh > max_oh) else "PART")
        #hit_type = "FULL" if (hmm_loh <= max_oh and hmm_roh <= max_oh) else ("OVER" if (left_oh <= max_oh and right_oh <= max_oh) else "PART")

        mean_score = 100.0 * hit.score / hit.aligned_cols
        mean_mcs   = 100.0 * hit.mcscore / hit.aligned_cols # hit.well_aligned_cols
        mean_wcs   = 100.0 * hit.wcscore / hit.aligned_cols # hit.well_aligned_cols
        identity   = 100.0 * hit.matches / hit.ali_len

        record = [ hit.hmm_name, str(hmm_len),
                   seq_name, str(seq_len), seqDict[seq_name]['Family'], seqDict[seq_name]['Function'],
                   "{:.2f}".format(hit.score),   "{:.4f}".format(mean_score),
                   "{:.2f}".format(hit.mcscore), "{:.4f}".format(mean_mcs),
                   "{:.2f}".format(hit.wcscore), "{:.4f}".format(mean_wcs),
                   "{:.2f}".format(identity),    "{:.2f}".format(hmm_cov), hit_type ]

        prev_hit = hit

    if len(record) > 0:
        print('\t'.join([ str(r) for r in record ]))

    return 0




# Check if the program is not being imported
if __name__ == "__main__":
    sys.exit(main())

